/*
 * Copyright 2007 Mark-Christoph Mller
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. */

package org.eml.MMAX2.annotation.markables;

import java.awt.Color;
import java.awt.Point;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.StringTokenizer;

import org.eml.MMAX2.api.MarkableRelationAPI;
import org.eml.MMAX2.core.MMAX2;

public class MarkableRelation implements MarkableRelationAPI
{
    // The name of the attribute the Markables of which this MarkableRelation handles (e.g. coref_class, antecedent, etc.)
    private String attributeName;
    // Whether individual MarkableSets or MarkablePointers (with maxsize>1) are ordered or unordered
    private boolean ordered;
    // Collection of all MarkableSets/MarkablePointers/SetPointers available for this attribute, mapped to their String values 
    private HashMap individualSets;
    /** Width of the line drawn when rendering sets. */
    private int lineWidth;
    
    /** Color of the line drawn used when rendering sets. */
    private Color lineColor;
    private Color defaultLineColor;
    /** Style of lines used when rendering sets: MMAX2.STRAIGHT or MMAX2.RCURVE or MMAX2.LCURVE. */
    private int lineStyle;        
    private int defaultLineStyle;
    
    private String attributeNameToShowInFlag="";
    
    private int maxSize;
    private boolean dashed;
    public MarkableRelation(String _attributeName, int type, boolean _ordered, int _lineWidth, Color _color, int _lineStyle, int _maxSize, boolean _dashed, String _attributeNameToShowInFlag)
    {
        attributeNameToShowInFlag = _attributeNameToShowInFlag;
        attributeName = _attributeName;
        ordered = _ordered;
        individualSets = new HashMap();
        lineWidth = _lineWidth;
        defaultLineColor = _color;        
        lineColor = _color;
        lineStyle = _lineStyle;
        defaultLineStyle = _lineStyle;
        maxSize = _maxSize;
        dashed = _dashed;
    }
    
    public final Color getLineColor()
    {
        return lineColor;
    }
    
    public final void setLineColor(Color _color)
    {
        if (_color == null)
        {
            lineColor = defaultLineColor;
        }
        else
        {
            lineColor = _color;
        }
    }

    public final String getAttributeNameToShowInFlag()
    {
        return attributeNameToShowInFlag;
    }
    
    public final int getLineStyle()
    {
        return lineStyle;
    }
    
    public final void setLineStyle(Integer _style)
    {
        if (_style == null)
        {
            lineStyle = defaultLineStyle;
        }
        else
        {
            lineStyle = _style.intValue();
        }
    }
    
    
    public final static Point calculateControlPoint(int x_source, int y_source, int x_target, int y_target, int lineStyle)
    {
        Point result = null;
        int x_diff=0;
        int y_diff=0;
        double x_ortho=0;
        double y_ortho=0;
        int x_middle=0;
        
        double c=0.2;
        
        if (lineStyle == MMAX2.XCURVE)
        {
            x_diff = x_target - x_source; 
            y_diff = y_target - y_source; 

            x_ortho =   y_diff * c; // * length_diff; 
            y_ortho = - x_diff * c; // * length_diff; 

            x_middle = (x_source + x_target)/2;  

            result = new Point((int)(x_middle + x_ortho),(int) (x_middle + y_ortho));
        }
        
        // Links oben nach rechts unten
        else if (x_source < x_target && y_source < y_target)
        {
            if (lineStyle == MMAX2.LCURVE)
            {
                result = new Point((int)x_target,(int) y_source);
            }
            else if (lineStyle == MMAX2.RCURVE)
            {
                result = new Point((int)x_source,(int) y_target);
            }
        }
        // Rechts oben nach links unten
        else if (x_source > x_target && y_source < y_target)
        {
            if (lineStyle == MMAX2.LCURVE)
            {
                result = new Point((int)x_source,(int) y_target);
            }
            else if (lineStyle == MMAX2.RCURVE)
            {
                result = new Point((int)x_target,(int) y_source);
            }
        }
        // Rechts unten nach links oben
        else if (x_source > x_target && y_source > y_target)
        {
            if (lineStyle == MMAX2.LCURVE)
            {
                result = new Point((int)x_target,(int) y_source);
            }
            else if (lineStyle == MMAX2.RCURVE)
            {
                result = new Point((int)x_source,(int) y_target);
            }
        }
        // Links unten nach rechts oben
        else if (x_source < x_target && y_source > y_target)
        {
            if (lineStyle == MMAX2.LCURVE)
            {
                result = new Point((int)x_source,(int) y_target);
            }
            else if (lineStyle == MMAX2.RCURVE)
            {
                result = new Point((int)x_target,(int) y_source);
            }
        }
        // Parallel
        else if (y_source == y_target)
        {
            if (x_target < x_source)
            {
                if (lineStyle == MMAX2.LCURVE)
                {
                    result = new Point((int)((int)x_source-(int)x_target)/2+(int)x_target,(int)y_source+20);
                }
                else if (lineStyle == MMAX2.RCURVE)
                {
                    result = new Point((int)((int)x_source-(int)x_target)/2+(int)x_target,(int)y_source-20);
                }
            }
            else
            {
                if (lineStyle == MMAX2.LCURVE)
                {
                    result = new Point((int)((int)x_target-(int)x_source)/2+(int)x_source,(int)y_source-20);
                }
                else if (lineStyle == MMAX2.RCURVE)
                {
                    result = new Point((int)((int)x_target-(int)x_source)/2+(int)x_source,(int)y_source+20);
                }
            }
        }
        // Senkrecht
        else if (x_source == x_target)
        {
            if (y_target < y_source)
            {
                if (lineStyle == MMAX2.LCURVE)
                {
                    result = new Point((int)x_source-20,(int)((int)y_source-(int)y_target)/2+(int)y_target);
                }
                else if (lineStyle == MMAX2.RCURVE)
                {
                    result = new Point((int)x_source+20,(int)((int)y_source-(int)y_target)/2+(int)y_target);
                }
            }
            else
            {
                if (lineStyle == MMAX2.LCURVE)
                {
                    result = new Point((int)x_source-20,(int)((int)y_target-(int)y_source)/2+(int)y_source);
                }
                else if (lineStyle == MMAX2.RCURVE)
                {
                    result = new Point((int)x_source+20,(int)((int)y_target-(int)y_source)/2+(int)y_source);
                }
            }            
        }
        return result;
    }

    public final MarkableSet[] getMarkableSets(boolean order)
    {
        ArrayList allSets = new ArrayList(individualSets.values());
        MarkableSet[] sets = (MarkableSet[])allSets.toArray(new MarkableSet[0]);
        if (order)
        {
            java.util.Arrays.sort(sets,new MarkableSetComparator());
        }
        
        return sets;
    }

    public final MarkablePointer[] getMarkablePointers(boolean order)
    {
        ArrayList allPointers = new ArrayList(individualSets.values());
        MarkablePointer[] pointers = (MarkablePointer[])allPointers.toArray(new MarkablePointer[0]);
        if (order)
        {
            java.util.Arrays.sort(pointers,new MarkablePointerComparator());
        }
        
        return pointers;
    }
    
    
    public final String getAttributeName()
    {
        return attributeName;
    }        
    
    /** This method adds Markable _markable to the MarkableSet attributeValue value (e.g. 'set_5'). If no set exists, a new one will be
        created. This method is called to add a Markable that has a valid value in the MARKABLE_SET-type attribute that this 
        MarkableRelation object is associated with. */
    public final MarkableSet addMarkableWithAttributeValueToMarkableSet(Markable _markable, String value)
    {
        // Try to get the MarkableSet already exisiting for value value
        MarkableSet oldSet = (MarkableSet)individualSets.get(value);
        if (oldSet == null)
        {
            oldSet = new MarkableSet(value, ordered, lineWidth, lineColor, lineStyle,this);
            individualSets.put(new String(value), oldSet);
        }
        // Now, oldSet is guaranteed to be a reference to a valid MarkableSet
        // Add Markable to be added
        oldSet.addMarkable(_markable);
        return oldSet;
    }
    
    public final MarkableSet getMarkableSetWithAttributeValue(String _attributeValue)
    {
        return (MarkableSet) individualSets.get(_attributeValue);
    }
    
    public final MarkableSet getMarkableSetContainingMarkable(Markable contained)
    {
        MarkableSet result = null;
        Iterator allMarkableSets = individualSets.values().iterator();
        while(allMarkableSets.hasNext())
        {
            result = (MarkableSet)allMarkableSets.next();
            if (result.containsMarkable(contained))
            {
                break;
            }
            else
            {
                result = null;
            }
        }
        return result;
    }
    
    public final void removeMarkableSet(MarkableSet removee)
    {
        individualSets.remove(removee.getAttributeValue());
    }            
    
    public final void removeMarkablePointer(MarkablePointer removee)
    {       
        individualSets.remove(removee.getSourceMarkable().getID());
    }            
    
    
    /** This method is used to create a new MarkablePointer with sourceMarkable as source Markable. The value that the source Markable
        has in its this.attributeName attribute is parsed into a list of satellite Markable IDs, which are then added to the set. 
        This method is called to add a Markable with a valid value in its MARKABLE_POINTER-type attribute. */
    public final void createMarkablePointer(Markable sourceMarkable, MarkableLevel sourceMarkableLevel)
    {
        MarkablePointer newMarkablePointer = new MarkablePointer(sourceMarkable,lineWidth,lineColor,lineStyle,maxSize,this,dashed);
        ArrayList allTargetIDs = parseTargetSpan(sourceMarkable.getAttributeValue(attributeName),";");
        for(int z=0;z<allTargetIDs.size();z++)
        {
            String currentIDEntry = (String)allTargetIDs.get(z);
            // Use full as default (assume no namespace)
            String currentID=currentIDEntry;
            // By default, assume target to be source
            String currentTargetLevelName=sourceMarkableLevel.getMarkableLevelName();
            MarkableLevel currentTargetMarkableLevel = sourceMarkableLevel;
            if (currentIDEntry.indexOf(":")!=-1)
            {
                // If the target id has a name space
                // Get that name space as new target levle name
                currentTargetLevelName = currentIDEntry.substring(0,currentIDEntry.indexOf(":"));
                currentTargetMarkableLevel = sourceMarkableLevel.getCurrentDiscourse().getMarkableLevelByName(currentTargetLevelName, false);
                currentID=currentIDEntry.substring(currentIDEntry.indexOf(":")+1);
            }
            
            
            Markable currentTarget = sourceMarkableLevel.getCurrentDiscourse().getCurrentMarkableChart().getMarkableByID(currentID,currentTargetMarkableLevel);
            //Markable currentSatellite = (Markable) sourceMarkableLevel.getMarkableById((String)allSatelliteIDs.get(z));
            if (currentTarget != null)
            {
                newMarkablePointer.addTargetMarkable(currentTarget);
            }
            else
            {
                System.err.println("Markable "+currentID+" was not found");
            }
        }
        individualSets.put(new String(sourceMarkable.getID()),newMarkablePointer);        
    }

        
    /** This method is used to create a new MarkablePointer with sourceMarkable as source Markable. The value that the source Markable
        has in its this.attributeName attribute is parsed into a list of satellite Markable IDs, which are then added to the set. 
        This method is called to add a Markable with a valid value in its MARKABLE_POINTER-type attribute. */
    public final void createMarkablePointer(Markable sourceMarkable, Markable firstTargetMarkable)
    {
        MarkablePointer newMarkablePointer = new MarkablePointer(sourceMarkable,lineWidth,lineColor,lineStyle,maxSize,this,dashed);
        newMarkablePointer.addTargetMarkable(firstTargetMarkable);
        individualSets.put(new String(sourceMarkable.getID()),newMarkablePointer);        
    }
    
    
    public final MarkablePointer getMarkablePointerForSourceMarkable(Markable sourceMarkable)
    {
        return (MarkablePointer) individualSets.get((String)sourceMarkable.getID());
    }
    
    public final MarkablePointer[] getMarkablePointersWithTargetMarkable(Markable target)
    {
        ArrayList result = new ArrayList();
        Iterator allPointerSets = individualSets.keySet().iterator();
        while (allPointerSets.hasNext())
        {
            MarkablePointer currentMarkablePointer = (MarkablePointer) individualSets.get((String)allPointerSets.next());
            if (currentMarkablePointer.isTargetMarkable(target))
            {
                result.add(currentMarkablePointer);
            }
        }
        return (MarkablePointer[]) result.toArray(new MarkablePointer[0]);
    }
    
    public final static String condenseSatelliteSpan_bak(String span)
    {
        String result = "";
        ArrayList list = parseTargetSpan(span,";");        
        for (int b=0;b<list.size();b++)
        {
            String currentEntry = (String)list.get(b);
            currentEntry = currentEntry.substring(currentEntry.indexOf("_")+1);
            if (b==0)
            {
                result = currentEntry;
            }
            else
            {
                result = result+" "+currentEntry;
            }
        }
        return result.trim();
    }

    /** This method accepts a target span string and returns a condensed form of it for display in the AttributeWindow. */
    public final static String condenseTargetSpan(String span)
    {
        String result = "";
        ArrayList list = parseTargetSpan(span,";");        
        for (int b=0;b<list.size();b++)
        {
            String currentEntry = (String)list.get(b);
            String currentNameSpace="";
            if (currentEntry.indexOf(":")!=-1)
            {
                // The current entry has a name space
                currentNameSpace = currentEntry.substring(0,currentEntry.indexOf(":")+1);
            }
            currentEntry = currentEntry.substring(currentEntry.indexOf("_")+1);
            if (b==0)
            {
                result = currentNameSpace+currentEntry;
            }
            else
            {
                result = result+" "+currentNameSpace+currentEntry;
            }
        }
        return result.trim();
    }
    
    public final static String expandTargetSpan(String span)
    {
        String result =MMAX2.defaultRelationValue;
        ArrayList list = parseTargetSpan(span," ");
        for (int z=0;z<list.size();z++)
        {            
            String currentEntry = (String)list.get(z);
            if (currentEntry.equals(MMAX2.defaultRelationValue))
            {
                break;
            }
            String currentNameSpace="";
            if (currentEntry.indexOf(":")!=-1)
            {
                // The current entry has a name space
                currentNameSpace=currentEntry.substring(0,currentEntry.indexOf(":")+1);
                currentEntry = currentEntry.substring(currentEntry.indexOf(":")+1);
            }
            if (z==0)
            {
                result = currentNameSpace+"markable_"+currentEntry;
            }
            else
            {
                result = result+";"+currentNameSpace+"markable_"+currentEntry;
            }
        }
        return result;
    }
    
    
    public final static ArrayList parseTargetSpan(String span, String delimiter)
    {
        ArrayList result = new ArrayList();
        StringTokenizer toki = new StringTokenizer(span,delimiter);
        while (toki.hasMoreTokens())
        {
            result.add(toki.nextToken());
        }        
        return result;
    }
    
    
    class MarkableSetComparator implements java.util.Comparator
    {        
        public int compare(Object o1, Object o2) 
        {
            MarkableSet s1 = (MarkableSet) o1;
            MarkableSet s2 = (MarkableSet) o2;
            int m1 = s1.getInitialMarkable().getLeftmostDiscoursePosition();
            int m2 = s2.getInitialMarkable().getLeftmostDiscoursePosition();
            if (m1 < m2) 
            {
                return -1;
            }
            else if (m2 < m1)
            {
                return 1;
            }
            else
            {
                return 0;
            }
        }        
    }
    
    class MarkablePointerComparator implements java.util.Comparator
    {        
        public int compare(Object o1, Object o2) 
        {
            MarkablePointer p1 = (MarkablePointer) o1;
            MarkablePointer p2 = (MarkablePointer) o2;
            int m1 = p1.getSourceMarkable().getLeftmostDiscoursePosition();
            int m2 = p2.getSourceMarkable().getLeftmostDiscoursePosition();
            if (m1 < m2) 
            {
                return -1;
            }
            else if (m2 < m1)
            {
                return 1;
            }
            else
            {
                return 0;
            }
        }        
    }    
}